/* The Ketchup program
 * A program to efficiently get recent logs
 * Copyright (C) 2001
 * Joshua Hill <josh-ketchup@untruth.org>
 * License: GPL
 */

/* This set of functions deals with checking filenames, paths and
 * directories for the correct permissions, ownership, etc.
 * WARNING: THESE FUNCTIONS ARE ROOT RUN!!!
 */

#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <errno.h>


#include "checks.h"
#include "util.h"
#include "setup.h"
#include "config.h"


/* This attempt to prevent the user from linking from 'bad files'
 * it:
 * Checks to see if the filename is too long
 * Checks to see if the file exists
 * Checks to see if the file is a normal file
 * checks to see if the file is SUID
 * checks to see if the file is SGID (not mandatory locking)
 * checks to see if the file is readable
 *
 * return value:
 *  function returns 1 if the file is ok
 *  function returns 0 if the file is not ok (and sets errno)
 */

int isGoodFile(const char *fileName) {
	int result;
	struct stat fileStats;
	int fd;

	/*Is the file name too long?*/
	if(my_strnlen(fileName, MAXFILENAMELENGTH + 1) > MAXFILENAMELENGTH) {
		if(config.isVerbose)
			fprintf(stderr, "Filename is too long\n");
		errno = ENAMETOOLONG;
		return 0;
	}

	/*file must exist*/
	result = stat(fileName, &fileStats);
	if(result < 0) {
		if(errno == ENOENT) {
			if(config.isVerbose)
				fprintf(stderr, "Filename does not exist\n");
		} else {
			perror("Couldn't stat file for permissions check");
		}
		return 0;
	}

	/*file must be a file*/
	if(!S_ISREG(fileStats.st_mode)) {
		if(config.isVerbose)
			fprintf(stderr, "File is not a normal file\n");
		errno = EACCES;
		return 0;
	}

	/*can't be SUID*/
	if(fileStats.st_mode & S_ISUID) {
		if(config.isVerbose)
			fprintf(stderr, "File is set-uid\n");
		errno = EACCES;
		return 0;
	}

	/*can't be SGID (must have group execute too, otherwise
	 *it's doing mandatory file locking...)*/
	if((fileStats.st_mode & (S_ISGID | S_IXGRP)) == (S_ISGID | S_IXGRP)) {
		if(config.isVerbose)
			fprintf(stderr, "File is set-gid\n");
		errno = EACCES;
		return 0;
	}

	/*file must be readable...*/
	fd = open(fileName, O_RDONLY);

	if(fd < 0) {
		if(config.isVerbose)
			fprintf(stderr, "File is not readable by current user\n");
		errno = EACCES;
		return 0;
	} else {
		result = close(fd);
		if(result < 0) {
			perror("Couldn't not close file for use");
			return 0;
		}
	}

	return 1;
}

/* This function attempts to enforce some basic directory restrictions,
 * so that the code can be consistent after stating a directory....
 *
 * return value:
 *  function returns 1 if the directory is ok
 *  function returns 0 if the directory is not ok, and sets errno.
 */
int isGoodDir(const char *dirName) {
	struct stat fileStats;
	int result;

	/*Is the directory name too long?*/
	if(my_strnlen(dirName, MAXFILENAMELENGTH + 1) > MAXFILENAMELENGTH) {
		if(config.isVerbose)
			fprintf(stderr, "Directory name is too long\n");
		errno = ENAMETOOLONG;
		return 0;
	}

	/*Now perform the tests on the directory...*/
	result = stat(dirName, &fileStats);

	/*Does it exist / is it accessible?*/
	if(result < 0) {
		if(errno == ENOENT) {
			if(config.isVerbose)
				fprintf(stderr, "directory does not exist\n");
		} else {
			perror("Could not stat directory for permissions check");
		}

		return 0;
	}

	/*The directory must be owned by the current real user*/
	if(fileStats.st_uid != getuid()) {
		if(config.isVerbose)
			fprintf(stderr, "Directory is not owned by current user...\n");
		errno = EACCES;
		return 0;
	}

	/*Must have no group or other read, write or execute, and cannot have sticky bit set*/
	if(fileStats.st_mode & 01077) {
		if(config.isVerbose)
			fprintf(stderr, "Directory has improper permissions\n");
		errno = EACCES;
		return 0;
	}

	/*The directory must be a directory (DUH!)*/
	if(!S_ISDIR(fileStats.st_mode)) {
		if(config.isVerbose)
			fprintf(stderr, "Directory is not a directory\n");
		errno = EACCES;
		return 0;
	}

	return 1;
}

/* This function attempts to see if the destination path for a link is
 * appropriate.
 */
int isGoodDestPath(const char *fileName) {
	char directory[MAXFILENAMELENGTH + 1];
	int cntr, cntr2;
	int length;

	length = my_strnlen(fileName, MAXFILENAMELENGTH+1);
	/*Is the directory name too long?*/

	if(length > MAXFILENAMELENGTH) {
		if(config.isVerbose)
			fprintf(stderr, "Path name is too long\n");
		errno = ENAMETOOLONG;
		return 0;
	}

	/*Find the last '/' character*/
	for(cntr = length - 1; cntr >= 0; cntr--) {
		if(fileName[cntr] == '/')
			break;
	}

	if(cntr == -1) {
		/*It appears to be just a filename...*/
		strcpy(directory, ".");
	} else {

		/*There's a directory in there, somewhere...*/
		for(cntr2 = 0; cntr2 <= cntr; cntr2++)
			directory[cntr2] = fileName[cntr2];

		assert(cntr2 <= MAXFILENAMELENGTH);

		directory[cntr2] = 0;
	}

	/*Now perform the tests on the directory...*/
	if(!isGoodDir(directory)) {
		return 0;
	}

	return 1;
}
