/* The Ketchup program
 * A program to efficiently get recent logs
 * Copyright (C) 2001
 * Joshua Hill <josh-ketchup@untruth.org>
 * License: GPL
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "state.h"
#include "util.h"
#include "ketchup.h"
#include "setup.h"
#include "config.h"

/*Global (shudder) config options.*/
struct configOptions config;

/*The main function
 * It:
 * parses the command line
 * opens the output file (if necessary)
 * Calls processFile for each file name
 */
int main(int argc, char *argv[]) {

	int option;

	char *stateDirName = NULL;
	char *outputFileName = NULL;
	int cntr=0;
	FILE *outputFile;
	int result;
	int expertMode = 0;

	/*Set up the process umask*/
	umask(077);

	/*Init the config structure*/
	config.isVerbose = 0;
	config.isSetGID = 0;
	config.savedGID = -1;

#ifdef CONFIG_SECURE_LINK
	/*Test to see if we are running set-gid to some other group, and if so
	 *remember what the id is supposed to be.  (for later use with the 
	 *helper program)
	 *See helperLink() and the README file for more details
	 */
	if(getgid() != getegid()) {
		fprintf(stderr, "Dealing with set-gid\n");
		config.isSetGID = 1;
		config.savedGID = getegid();

		/*Now restore the normal user's gid*/
		result = setegid(getgid());

		if(result < 0) {
			perror("Error doing the dance of the GIDs");
			exit(-2);
		}
	} else {
		config.isSetGID = 0;
	}
#endif
	/*Parse the command line*/
	while((option = getopt(argc, argv, "s:o:evh")) != -1) {
		switch (option) {
			case 's':
				stateDirName = strdup(optarg);
				if(!stateDirName) {
					perror("Couldn't duplicate state directory name");
					exit(-1);
				}
				break;
			case 'o':
				if(!strncmp(optarg, "-", 1))
					break;
				else
					outputFileName = strdup(optarg);
					if(!outputFileName) {
						perror("Couldn't duplicate output file name");
						exit(-1);
					}
					break;
			case 'e':
				expertMode = 1;
				break;
			case 'v':
				config.isVerbose = 1;
				break;
			case 'h':
			default:
				usage();
				exit(-1);
			}
	}

	/*Grab local copies of user input*/
	if(!stateDirName) {
		stateDirName = strdup(".ketchup");
		if(!stateDirName) {
			perror("Couldn't duplicate state directory name");
			exit(-1);
		}
	}

	if(outputFileName) {
		if((outputFile = fopen(outputFileName, "w")) == NULL) {
			perror("Can't open output file");
			exit(-1);
		}
	} else {
		outputFile = stdout;
	}

	/*Now iterate through the specified files and dump changes*/
	if(config.isVerbose)
		fprintf(stderr, "Starting to process files...\n");

	for(cntr=optind; cntr < argc; cntr++) {
		if(!expertMode && !isAbsolutePath(argv[cntr])) {
			fprintf(stderr, "Error: Data filename (%s) should be absolute not relative\n", argv[cntr]);
		} else {
			if(optind != (argc - 1)) {
				printf("---[%s]---\n", argv[cntr]);
			}

			result = processFile(argv[cntr], stateDirName, outputFile);
			if(result < 0) {
				fprintf(stderr, "Error when dealing with %s\n", argv[cntr]);
			}
		}
	}

	/*Clean up...*/
	if(stateDirName)
		free(stateDirName);

	if(outputFileName)
		free(outputFileName);

	return 0;
}

/* This function does the processing for one file
 * It:
 * opens the state directory (openStateDataDir)
 * outputs new information (outputNewData)
 * closes the state directory (closeStateDataDir)
 */
int processFile(char* dataFileName, char *stateDirName, FILE *outputFile) {

	int result;
	struct stateDirHandle *fileState;
	int processResult = 0;

	assert(dataFileName);
	assert(stateDirName);
	assert(outputFile);

	if((fileState = openStateDataDir(stateDirName, dataFileName)) == NULL) {
		fprintf(stderr, "Can't open state data directory\n");
		return -1;
	} 

	/*Output all the new info now.
	 *Note that even if there is an error with the function, 
	 *closeStateDataDir is called, anyway to cleanup...*/

	result = outputNewData(fileState, outputFile);
	if(result < 0) {
		fprintf(stderr, "Couldn't output requested data\n");
		processResult = -1;
	}

	/*Now close the state data directory*/
	result = closeStateDataDir(fileState);
	if(result < 0) {
		fprintf(stderr, "Error while closing state data directory\n");
		processResult = -1;
	}

	return processResult;
}

/* This function outputs the new lines in the files
 * The main workhorse function
 * It:
 * goes to the stored offset
 * gets complete lines, and outputs them
 * gets the current position (end of the last output line)
 * sets the stored offset
 * if the local and real data files are different files then
 *   finishes the file out (even if it is not a complete line)
 *   refreshes the state to the new file (refreshState)
 *   if refreshState indicates that we have a new file then
 *     Call outputNewData again for the new file (max 1 level of recursion)
 *   endif
 * endif
 *
 * Notes:
 * Why is curLine size MAXLINELENGTH + 2?  Well, we want to be able
 * to deal with MAXLINELENGTH user data bytes, so we need 1 for the
 * NULL termination.  In addition to that, we need the ability to
 * deal with long lines by just appending a LF at the end.
 * So, really, you could end up with a line that is MAXLINELENGTH
 * non-LF characters, and then an (ketchup appended) LF.
 */
int outputNewData(struct stateDirHandle *fileState, FILE *output) {
	int result;
	long curPos;

	char curLine[MAXLINELENGTH + 2]; 
	char *strResult;

	assert(fileState);
	assert(output);

	assert(fileState->localDataFile);
	assert(fileState->filePos >= 0);
	assert(fileState->localDataFileInode != 0);
	assert(fileState->dataFileInode != 0);
        


	result = fseek(fileState->localDataFile, fileState->filePos, SEEK_SET);
	if(result < 0) {
		perror("Can't seek in local dataFile");
		return -1;
	}

	strResult = getLine(curLine, MAXLINELENGTH + 2, fileState->localDataFile);

	while(strResult != NULL){
		result = fputs(curLine, output);
		if(result == EOF) {
			perror("Can't output file line");
			return -1;
		}
		strResult = getLine(curLine, MAXLINELENGTH + 2, fileState->localDataFile);
	}

	curPos = ftell(fileState->localDataFile);
	if(curPos < 0) {
		perror("Can't get file offset");
		return -1;
	}

	fileState->filePos = curPos;

	if(fileState->localDataFileInode != fileState->dataFileInode) {
		/*the files have changed...
		 *first finish off the file*/
		strResult = finishFile(curLine, MAXLINELENGTH + 2, fileState->localDataFile);
		if(strResult) {
			result = fputs(curLine, output);
			if(result == EOF) {
				perror("Can't output file line");
				return -1;
			}
		}

		/*now refresh the file*/
		result = refreshState(fileState);
		if(result < 0) {
			fprintf(stderr, "Couldn't refresh the state\n");
			return -1;
		}

		if(result > 0) {
			/*OK, this should be the case, but avoid infinite recursions...*/
			assert(fileState->localDataFileInode == fileState->dataFileInode);

			/*Recurse, max one level*/
			result = outputNewData(fileState, output);
			if(result < 0) {
				fprintf(stderr, "Couldn't output new file's state\n");
				return -1;
			}
		}
	}

	return 0;
}

void usage() {
	printf("Ketchup version %s\n", VERSION);
	printf("Usage: ketchup [OPTION]... FILENAME...\n");
	printf("Report then additional (appended) lines in FILENAME since the last\n");
	printf("invocation of ketchup.  If ketchup has never been called for this\n");
	printf("filename, output all complete lines in FILENAME.\n");
	printf("Options:\n");
	printf("  -s STATEDIR\t\tSpecify a location for the data state directories\n");
	printf("  -o FILENAME\t\tOutput data to FILENAME, instead of STDOUT\n");
	printf("  -e\t\t\tExpert mode.  Don't complain about relative pathnames\n");
	printf("  -v\t\t\tVerbose mode.  Output more status and error information\n");
	printf("  -h\t\t\tHelp.  Display this output\n");
}
