/* klink, a privileged helper, meant to be run as suid-root <sharp inhale>.
 * This helper may be required on systems that use Solar Designer's openwall 
 * linux kernel patch, and make use of the CONFIG_SECURE_LINK feature of 
 * this patch (as the author of this software does on his production 
 * machines). This helper is supposed to be secure, and is designed to 
 * only allow users to hard link a file.  Assuming this is the case 
 * (<titter>), users who can run this should get (worst case) restricted 
 * use of the normal link(2) function.  Of course, the fact that you're 
 * running the CONFIG_SECURE_LINK option in the first place probably means 
 * that you think that hard links should be restricted.  This program 
 * attempts to enforce (most of) the restrictions imposed by Solar Designer's 
 * CONFIG_SECURE_LINK patch.  
 *
 * This wrapper attempts to run in the least privileged way it can.
 * The wrapper should be suid root, but the wrapper discards all
 * rootly powers except for CAP_FOWNER and CAP_DAC_OVERRIDE.  NOTE:
 * This is more privileged than I would like (CAP_DAC_OVERRIDE is
 * ridiculously powerful, but if I don't include it, root can't
 * create a file in the directory... <grumble> <grumble>)
 * The wrapper runs as the executing user for most of
 * its existence, but must switch back to the (reduced) root in
 * order to actually make the link.
 *
 * This program:
 * restricts capabilities to CAP_FOWNER and CAP_DAC_OVERRIDE only
 * drops root (must keep it saved, or the caps are cleared)
 * parses command line
 * checks for file based source 
 * checks the destination directory (for the sticky bit, ownership,
 *  permissions, directory-ness, etc)
 * changes euid to 0 (get CAP_FOWNER and CAP_DAC_OVERRIDE)
 * does the link
 * drops root (drop CAP_FOWNER and CAP_DAC_OVERRIDE)
 * checks the destination file
 * if destination file is bad then
 *   unlink the file
 * endif
 *
 * WARNING: THESE FUNCTIONS ARE ROOT RUN!!!
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>

#include <sys/capability.h>
#include "setup.h"
#include "klink.h"
#include "checks.h"
#include "config.h"

/*Global (shudder) config options.*/
struct configOptions config;

int main(int argc, char *argv[]) {

	uid_t ruid, euid;
	int result;
	int returnValue;

	extern char **environ;

	returnValue = 0;

	/*Init config structure.  We don't parse a command line really, 
	 *so these never change*/
	config.isVerbose = 1;
	config.isSetGID = 0;
	config.savedGID = -1;

	/*look at RUID/EUID and make sure that we're root...*/
	ruid = getuid();
	euid = geteuid();

	if(euid != 0) {
		fprintf(stderr, "wrapper must be run as root\n");
		exit(1);
	}

	/*Discard unused caps*/
	result = discardCaps();

	if(result < 0) {
		fprintf(stderr, "Couldn't discard capabilities\n");
		exit(1);
	}

	/*Is this an SUID binary?*/
	if(ruid != euid) {
		/*set the euid to the normal user*/
		result = seteuid(ruid);

		if(result < 0) {
			perror("Couldn't set euid");
			exit(1);
		}
	}

	/*Were we called correctly?*/

#ifndef DEBUG
	/*Check to see that the environment is empty*/
	/*(as ketchup sets it)*/
	if(environ[0] != NULL) {
		fprintf(stderr, "klink is a helper program, and should never be directly run by a user\n");
		exit(1);
	}
		
#endif
	
	/*Correct is klink <src> <dest>*/
	if(argc != 3) {
		fprintf(stderr, "Command line is incorrect.\n");
		exit(1);
	}

	/*Check the provided source and dest paths...*/
	if(!isGoodFile(argv[1])) {
		perror("Source file is inappropiate");
		exit(2);
	}

	if(!isGoodDestPath(argv[2])) {
		perror("Destination directory is inappropiate");
		exit(2);
	}

	/*if klink is an SUID program, switch back to the privileged user*/
	if(ruid != 0) {
		result = seteuid(euid);
		if(result < 0) {
			fprintf(stderr, "Can't restore permissions\n");
			exit(4);
		}
	}

	/*Do the link*/
	result = link(argv[1], argv[2]);
	if(result < 0) {
		perror("Can't create link");
		exit(3);
	}

	/*if klink is an SUID program, switch back to the non-privileged user*/
	if(ruid != 0) {
		result = seteuid(ruid);
		if(result < 0) {
			fprintf(stderr, "Can't restore permissions\n");
			returnValue = 4;
		}
	}

	/*Make sure that the actual destination was ok...*/
	if(!isGoodFile(argv[2]) || !isGoodDestPath(argv[2])) {
		/*Nope.  Try to clean up.*/
		result = unlink(argv[2]);
		fprintf(stderr, "Destination file or directory was inappropiate\n");
		fprintf(stderr, "RACE CONDITION\n");
		if(result < 0)
			fprintf(stderr, "Cleanup unsuccessful\n");

		exit(5);
	}

	return returnValue;
}

/* This function discards all capabilities but CAP_FOWNER and
 * CAP_DAC_OVERRIDE
 */

int discardCaps() {
	cap_t newCaps;
	cap_value_t capList[2];

	/*init a new cap set*/
	newCaps = cap_init();
	if(!newCaps) {
		perror("Can't create (empty) cap set");
		return -1;
	}

	/*Set the permitted caps*/
	capList[0] = CAP_FOWNER;
	capList[1] = CAP_DAC_OVERRIDE;
	cap_set_flag(newCaps, CAP_PERMITTED, 2, capList, CAP_SET);

	/*Set the effective cap*/
	capList[0] = CAP_FOWNER;
	capList[1] = CAP_DAC_OVERRIDE;
	cap_set_flag(newCaps, CAP_EFFECTIVE, 2, capList, CAP_SET);

	if(cap_set_proc(newCaps) != 0) {
		perror("Can't set caps");
		return -1;
	}

	return 0;
}

/* This is a debug function that prints out the current capability set
 * in a controlled manner.
 */
void printCaps() {
	cap_t process_caps;
	char *capString;
	ssize_t length;

	process_caps = cap_get_proc();
	if(process_caps) {
		capString = cap_to_text(process_caps, &length);
		if(capString) {
			fprintf(stderr, "%s\n", capString);
		} else {
			perror("Can't convert capabilities to a string");
			abort();
		}
	} else {
		perror("Can't get current process capabilities");
		abort();
	}
}

