/* The Ketchup program
 * A program to efficiently get recent logs
 * Copyright (C) 2001
 * Joshua Hill <josh-ketchup@untruth.org>
 * License: GPL
 */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include "checks.h"
#include "link.h"
#include "config.h"
#include "setup.h"

/*This function attempts to make systems that run Solar Designer's kernel
 *patch (with CONFIG_SECURE_LINK enabled) work, by using a privileged 
 *helper program to do the link.  
 *
 * It:
 * Checks the src and dst for good permissions
 * does the link
 * makes sure that the result (and dst path) still have good permissions
 */
int doLink(const char *oldpath, const char *newpath) {
	int result;
	int linkResult;

	if(!isGoodFile(oldpath)) {
		return -1;
	}

	if(!isGoodDestPath(newpath)) {
		return -1;
	}

#ifndef CONFIG_SECURE_LINK
	linkResult = link(oldpath, newpath);
#else
	linkResult = helperLink(oldpath, newpath);
#endif

	if(linkResult >= 0) {
		if(!isGoodFile(newpath) || !isGoodDestPath(newpath)) {
			result = unlink(newpath);
			if(result < 0) {
				perror("Couldn't unlink nasty-bad new link");
			}
			fprintf(stderr, "Naughty user! Race conditions are for kids!\n");
			return -1;
		}
	} else {
		perror("Couldn't establish link");
	}
			
	return linkResult;
}

/*This is the actual function that does all the messy fork/exec stuff
 *for the helper, and attempts to extract the return information from
 *helper program.  It also interacts with the saved GID that we noted
 *at the start of the program.  The child needs to run as that GID.
 *
 *It sets up the process's ARGV
 *sets the process's environment to empty (klink checks for this)
 *forks
 *The child then:
 *  if ketchup is SGID then
 *     restore the privliged GID
 *  endif
 *  execs the helper (klink)
 *
 * The parent then:
 *  waits for the child to exit
 *  parses the return value
 *  
 */
int helperLink(const char *oldpath, const char *newpath) {
	pid_t pid;
	char *argv[4];
	char *envp[1];
	int status;
	int result;
	int helperReturn;

	/*Set up the command line and the environment*/
	argv[0] = strdup(HELPERPATH);
	argv[1] = strdup(oldpath);
	argv[2] = strdup(newpath);
	argv[3] = NULL;

	if(!(argv[0] && argv[1] && argv[2])) {
		perror("Couldn't dup user strings");
		return -1;
	}

	envp[0] = NULL;

	pid = fork();

	if(pid < 0) {
		perror("error forking for helper");
		return -1;
	}

	/*Is it the child?*/
	if(pid == 0) {
		/*if the program was sgid, make the gid active*/
		if(config.isSetGID) {
			result = setegid(config.savedGID);
			if(result < 0) {
				perror("Couldn't restore gid");
				_exit(254);
			}
		}

		result = execve(HELPERPATH, argv, envp);
		if(result < 0) {
			perror("error execing helper");
			_exit(255);
		}
	}

	/*OK, it's the parent...*/
	/*Wait for the helper program to complete...*/
	result = waitpid(pid, &status, 0);

	if(result < 0) {
		perror("waiting for the child process failed");
		return -1;
	}

	/*Free the argv strings...*/
	free(argv[0]);
	free(argv[1]);
	free(argv[2]);

	/*parse the status information*/
	if(WIFEXITED(status)) {
		helperReturn = WEXITSTATUS(status);
		if(helperReturn == 0) {
			return 0; /*Everything was good...*/
		} else {
			switch(helperReturn) {
				case 1:
					fprintf(stderr, "helper init failure\n");
					break;
				case 2:
					fprintf(stderr, "Helper refused to make link.\n");
					break;
				case 3:
					fprintf(stderr, "Helper unexpectedly unable to make link\n");
					break;
				case 4:
					fprintf(stderr, "privilege handling error.\n");
					break;
				case 5:
					fprintf(stderr, "I SMELL AMBUSH!!!\n");
					break;
				case 254:
					fprintf(stderr, "Restore privileges to child\n");
					break;
				case 255:
					fprintf(stderr, "Could not start helper\n");
					break;
				default:
					fprintf(stderr, "What? Unknown error.\n");
					break;
			}

			errno = ENOEXEC;

			return -1;
		}
	} else {
		errno = EINTR;
		return -1;
	}
}
