/* The Ketchup program
 * A program to efficiently get recent logs
 * Copyright (C) 2001
 * Joshua Hill <josh-ketchup@untruth.org>
 * License: GPL
 */

#include <errno.h>
#include <assert.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/file.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "state.h"
#include "checks.h"
#include "util.h"
#include "setup.h"
#include "link.h"

/*This function opens the state directory.
 *It:
 *makes sure the state directory exists
 *makes sure that the state directory has the the right permissions
 *grabs the lock for the state directory
 *if the state data directory does not exist then
 *   create the state data directory
 *endif
 *creates the fileState structure and initializes it
 *fills in the fileState structure as much as possible without stored state
 *gets the stored state 
 *opens the local data file
 *sets the inodes for later comparison
 */
struct stateDirHandle *openStateDataDir(char *stateDirName, char *dataFileName) {

	struct stat fileStats;
	int result;
	char *baseDataFileName;
	char stateDirLockFileName[MAXFILENAMELENGTH + 1];
	char stateDataDirName[MAXFILENAMELENGTH + 1];
	struct stateDirHandle *fileState;
	int lockFile;

	assert(stateDirName);
	assert(dataFileName);

	if(!isGoodDir(stateDirName)) {
		fprintf(stderr, "State directory is invalid\n");
		return NULL;
	}

	/*Now grab the lock associated with this file*/
	baseDataFileName = getBaseName(dataFileName);
	if(baseDataFileName == NULL) {
		fprintf(stderr, "Can't get file's base name\n");
		return NULL;
	}

	result = snprintf(stateDirLockFileName, MAXFILENAMELENGTH, "%s/.%s.lock", stateDirName, baseDataFileName);

	if((result < 0) || (result >= sizeof(stateDirLockFileName))) {
		fprintf(stderr, "String too large...\n");
		return NULL;
	}

	/*grab the lock*/
	lockFile = lockStateDataDir(stateDirLockFileName);

	if(lockFile < 0) {
		fprintf(stderr, "Couldn't get lock for %s\n", dataFileName);
		return NULL;
	}

	/*check to see if the state data directory exists*/
	result = snprintf(stateDataDirName, MAXFILENAMELENGTH, "%s/%s", stateDirName, baseDataFileName);
	if((result < 0) || (result >= sizeof(stateDataDirName))) {
		fprintf(stderr, "String too large...\n");
		return NULL;
	}

	if(!isGoodDir(stateDataDirName)) {
		if(errno == ENOENT) {
			/*didn't exist, let's create it...*/
			errno = 0;
			result = initStateDataDir(stateDataDirName, dataFileName);
			if(result < 0) {
				fprintf(stderr, "Could not initialize state data\n");
				return NULL;
			}

			if(!isGoodDir(stateDataDirName)) {
				fprintf(stderr, "stat data dir was not appropiate after creation\n");
				return NULL;
			}
		} else {
			/*hmmm... other error...*/
			fprintf(stderr, "Couldn't stat the state data directory\n");
			return NULL;
		}
	}

	/*OK, now create the filestructure...*/
	fileState = (struct stateDirHandle *)malloc(sizeof(struct stateDirHandle));
	if(!fileState) {
		perror("Error getting fileState memory");
		return NULL;
	}

	/*Initialize the structure*/
	fileState->stateDirName = NULL;
	fileState->dataFileName = NULL;
	fileState->stateDirLockFileName = NULL;
	fileState->stateDataDirName = NULL;
	fileState->stateDirLockFile = 0;

	fileState->localDataFileName = NULL;
	fileState->stateFileName = NULL;
	fileState->localDataFile = NULL;

	fileState->localDataFileInode = 0;
	fileState->dataFileInode = 0;
	fileState->deleteMe = 0;

	/*Now fill it out...*/

	fileState->stateDirName = strdup(stateDirName);
	fileState->dataFileName = strdup(dataFileName);
	fileState->stateDirLockFileName = strdup(stateDirLockFileName);
	fileState->stateDataDirName = strdup(stateDataDirName);

	if(!(fileState->stateDirName && fileState->dataFileName && 
	 fileState->stateDirLockFileName && fileState->stateDataDirName)) {
		perror("Couldn't dupe config strings");
		return NULL;
	}

	fileState->stateDirLockFile = lockFile;

	result = readState(fileState);
	if(result < 0) {
		fprintf(stderr, "Could not read the state file\n");
		return NULL;
	}

	if((fileState->localDataFile = fopen(fileState->localDataFileName, "r")) == NULL) {
		perror("Can't open the local data file");
		delStateDataDir(fileState);
		return NULL;
	}

	/*Verify that the localDataFile is appropiate...*/
	if(!isGoodFile(fileState->localDataFileName)) {
		fprintf(stderr, "Local data file is not appropiate\n");
		return NULL;
	}

	/*Now get the inode...*/
	result = stat(fileState->localDataFileName, &fileStats);
	if(result < 0) {
		perror("Couldn't stat the localDataFile... again");
		return NULL;
	}

	fileState->localDataFileInode = fileStats.st_ino;

	/*Verify that the dataFile is appropriate*/
	if(!isGoodFile(fileState->dataFileName)) {
		/*Hmm... Something is amiss with the dataFile.
		 *Let's check to see if we can proceed, anyway...*/
		if(errno == ENOENT) {
			fprintf(stderr, "data file does not exist currently\n");
			fileState->dataFileInode = -1;
		} else {
			fprintf(stderr, "data file is a bad, bad file...\n");
			return NULL;
		}
	} else {
		/*Good file, now get the inode...*/
		result = stat(fileState->dataFileName, &fileStats);
		if(result < 0) {
			perror("Couldn't stat the dataFile... again");
			return NULL;
		}

		fileState->dataFileInode = fileStats.st_ino;
	}
	
	/*Verify that it was filled out...*/
	assert(fileState);
	assert(fileState->stateDirName);
	assert(fileState->dataFileName);
	assert(fileState->stateDirLockFileName);
	assert(fileState->stateDataDirName);
	assert(fileState->stateDirLockFile > 0);

	assert(fileState->localDataFileName);
	assert(fileState->stateFileName);
	assert(fileState->localDataFile);

	assert(fileState->localDataFileInode != 0);
	assert(fileState->dataFileInode != 0);

	assert(fileState->filePos >= 0);

	return fileState;
}

/* This function initializes a new state data directory
 * It:
 * makes the new directory
 * makes the local data file (hard link)
 * makes the state file
 * 
 * Note: the program attempts to not leave a partially initilized
 * state data directory...
 */
int initStateDataDir(const char *stateDirName, const char *dataFileName) {
	int result;
	char stateFileName[MAXFILENAMELENGTH + 1];
	char localDataFileName[MAXFILENAMELENGTH + 1];
	FILE *stateFile;


	assert(stateDirName);
	assert(dataFileName);

	/*Make the new directory*/
	result = mkdir(stateDirName, 0700);

	if(result < 0) {
		perror("Can't make state directory");
		return -1;
	}

	/*Make the new local data file*/
	result = snprintf(localDataFileName, MAXFILENAMELENGTH, "%s/%s", stateDirName, LOCALDATAFILENAME);
	if((result < 0) || (result >= sizeof(localDataFileName))) {
		fprintf(stderr, "String too large...\n");
		result = rmdir(stateDirName);
		if(result < 0) {
			perror("Couldn't clean up stateDataDir failure(1)");
		}
		return -1;
	}

	result = snprintf(stateFileName, MAXFILENAMELENGTH, "%s/%s", stateDirName, STATEFILENAME);
	if((result < 0) || (result >= sizeof(stateFileName))) {
		fprintf(stderr, "String too large...\n");
		result = rmdir(stateDirName);
		if(result < 0) {
			perror("Couldn't clean up stateDataDir failure(2)");
		}
		return -1;
	}

	result = doLink(dataFileName, localDataFileName);

	if(result < 0) {
		fprintf(stderr, "Can't link data file to local data file\n");
		result = rmdir(stateDirName);
		if(result < 0) {
			perror("Couldn't clean up stateDataDir failure(3)");
		}
		return -1;
	}

	/*Make the state file...*/
	if((stateFile = fopen(stateFileName, "w")) == NULL) {
		perror("Couldn't open state file\n");
		result = unlink(localDataFileName);
		if(result < 0) {
			perror("Couldn't clean up stateDataDir failure(4)");
		}
		result = rmdir(stateDirName);
		if(result < 0) {
			perror("Couldn't clean up stateDataDir failure(5)");
		}
		return -1;
	}

	fprintf(stateFile, "0\n");
	result = fclose(stateFile);
	/*Reset to flag value*/
	stateFile = NULL;

	if(result < 0) {
		perror("Couldn't close state file");
		result = unlink(stateFileName);
		if(result < 0) {
			perror("Couldn't clean up stateDataDir failure(6)");
		}
		result = unlink(localDataFileName);
		if(result < 0) {
			perror("Couldn't clean up stateDataDir failure(7)");
		}
		result = rmdir(stateDirName);
		if(result < 0) {
			perror("Couldn't clean up stateDataDir failure(8)");
		}
		return -1;
	}
	
	return 0;
}

/* This function obtains the lock file for a state data directory
 * It:
 * it opens the lock file
 * obtains an OS lock on the file (using flock)
 * truncates the file
 * writes the PID of the ketchup process to the file
 * returns the file descriptor
 */
int lockStateDataDir(const char *stateDirLockFileName) {
	int fd;
	int result;
	char pidstr[10];
	pid_t pid;

	assert(stateDirLockFileName);

	fd = open(stateDirLockFileName, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);

	if(fd < 0) {
		perror("Can't open lock file");
		return -1;
	}

	result = flock(fd, LOCK_EX|LOCK_NB);

	if(result < 0) {
		perror("Can't obtain lock");
		return -1;
	}

	result = ftruncate(fd, 0);

	if(result < 0) {
		perror("Can't truncate file");
	}

	pid = getpid();

	result = snprintf(pidstr, 10, "%d\n", pid);
	if((result < 0) || (result >= sizeof(pidstr))) {
		fprintf(stderr, "String too large...\n");
		return -1;
	}


	result = write(fd, pidstr, strlen(pidstr));

	if(result < 0) {
		perror("Couldn't write PID to file");
	}

	return fd;
}

/* This function unlocks the state data directory
 * It:
 * truncates the file (removes the PID -- obscure race) not unlink? see note.
 * unlock file
 * close the file
 *
 * Note: I would like to do an unlink here, but if I do that there could
 * be contention in the case where process 1 is holding the lock,
 * process 2 starts, opens the (locked) file and then process 1 unlocks
 * the file and unlinks.  Process 2 then tries to flock the file, it succeeds
 * and process 2 continues along.  Process 3 tries to look at the same file,
 * opens a new file (the old one was deleted), obtains a lock on the new
 * file and continues.  Now you have both process 2 and 3 operating at the
 * same time on the same files.  SUCK!
 */
int unlockStateDataDir(struct stateDirHandle *fileState) {
	int result;
	int unlockResult = 0;

	assert(fileState);
	assert(fileState->stateDirLockFile > 0);

	result = ftruncate(fileState->stateDirLockFile, 0);
	if(result < 0) {
		perror("Couldn't truncate the lockfile for close");
		unlockResult = -1;
	}

	result = flock(fileState->stateDirLockFile, LOCK_UN);
	if(result < 0) {
		perror("Couldn't unlock the lockfile");
		unlockResult = -1;
	}

	result = close(fileState->stateDirLockFile);
	/*reset to flag value*/
	fileState->stateDirLockFile = 0;

	if(result < 0) {
		perror("Couldn't close the lockfile");
		unlockResult = -1;
	}

	return unlockResult;
}

/* This function gets the data out of the state file
 * This data is currently only the previous file offset.
 * This is obviously not very interesting at the moment, however
 * this is one of the places to possibly be improved in the future...
 * It:
 * Generates the local data file name and fills in the field.
 * Generates the state file name and fills in the field.
 * Opens the state file
 * reads in the current offset
 * closes the state file
 */
int readState(struct stateDirHandle *fileState) {
	char temp[MAXFILENAMELENGTH + 1];
	FILE *stateFile;
	int result;

	assert(fileState);
	assert(fileState->stateDataDirName);

	result = snprintf(temp, MAXFILENAMELENGTH, "%s/%s", fileState->stateDataDirName, LOCALDATAFILENAME);
	if((result < 0) || (result >= sizeof(temp))) {
		fprintf(stderr, "String too large...\n");
		return -1;
	}

	fileState->localDataFileName = strdup(temp);
	if(!(fileState->localDataFileName)) {
		perror("Couldn't duplicate local Data File Name");
		return -1;
	}

	result = snprintf(temp, MAXFILENAMELENGTH, "%s/%s", fileState->stateDataDirName, STATEFILENAME);
	if((result < 0) || (result >= sizeof(temp))) {
		fprintf(stderr, "String too large...\n");
		return -1;
	}

	fileState->stateFileName = strdup(temp);
	if(!(fileState->stateFileName)) {
		perror("Couldn't duplicate State File Name");
		return -1;
	}

	stateFile = fopen(fileState->stateFileName, "r");
	if(stateFile != NULL) {
		fscanf(stateFile, "%ld\n", &(fileState->filePos));

		result = fclose(stateFile);
		/*reset to flag value...*/
		stateFile = NULL;

		if(result < 0) {
			perror("Could not close state file");
			return -1;
		}
	} else if(errno == ENOENT) {
		/*Special case: if the state file doesn't exist*/
		/*set the values to default values and proceed*/
		fileState->filePos = 0;
	} else {
		perror("Can't open the statefile for read");
		return -1;
	}

	return 0;
}

/* This file writes out the current state
 * This is the associated close function for readState, so if readState
 * gets more complex, this function will too...
 * It:
 * Opens the state file
 * writes out the current file offset
 * closes the state file
 */
int writeState(struct stateDirHandle *fileState) {
	FILE *stateFile;
	int result;

	assert(fileState);
	assert(fileState->stateFileName);
	assert(fileState->filePos >= 0);

	if((stateFile = fopen(fileState->stateFileName, "w")) == NULL) {
		perror("Can't open the statefile for write");
		return -1;
	}

	fprintf(stateFile, "%ld\n", fileState->filePos);

	result = fclose(stateFile);
	/*Reset to flag value*/
	stateFile = NULL;
	if(result < 0) {
		perror("Can't close the statefile");
		return -1;
	}

	return 0;
}

/* This function closes the state directory
 * It:
 * closes the local data file
 * if the directory should be deleted then
 *   delete it
 * else
 *   writes the state out (writeState)
 * endif
 * unlocks the state directory (unlockStateDataDir)
 * frees the now unused values in the fileState structure
 * frees the fileState structure
 */
int closeStateDataDir(struct stateDirHandle *fileState) {
	int result;
	int retValue;

	assert(fileState);
	assert(fileState->stateDirName);
	assert(fileState->dataFileName);
	assert(fileState->stateDirLockFileName);
	assert(fileState->stateDataDirName);
	assert(fileState->stateFileName);
	assert(fileState->localDataFileName);
	assert(fileState->localDataFile);
	
	retValue = 0;

	/*now close the localDataFile*/
	result = fclose(fileState->localDataFile);

	/*reset to flag value*/
	fileState->localDataFile = NULL;

	if(result < 0) {
		perror("Couldn't close local data file");
		retValue = -1;
	}

	if(fileState->deleteMe) {
		result = delStateDataDir(fileState);
		if(result < 0) {
			fprintf(stderr, "Error deleting the state directory\n");
			retValue = -1;
		}
	} else {
		result = writeState(fileState);

		if(result < 0) {
			fprintf(stderr, "Could not write state out\n");
			retValue = -1;
		}
	}

	result = unlockStateDataDir(fileState);

	if(result < 0) {
		fprintf(stderr, "Could not unlock state directory\n");
		retValue = -1;
	}

	free(fileState->stateDirName);
        free(fileState->dataFileName);
        free(fileState->stateDirLockFileName);
        free(fileState->stateDataDirName);
	free(fileState->stateFileName);
	free(fileState->localDataFileName);
	free(fileState);

	return retValue;
}

/* This file refreshes the current state by closing the localdata file,
 * deleting it, and then refreshing the link.  This is a similar process
 * to initStateDataDir.  This function needs to return a number less than 0
 * on error, 0 on no re-output required, at greater than 0 if the new
 * file needs to be output as well...
 * It:
 * closes the local data file
 * unlinks the local data file
 * links the data file to a new local data file
 * updates the file position
 * opens the new local data file
 * stats the new local data file
 * makes sure that it is the expected inode
 * if not the expected inode then
 *    output an error message
 *    update the data file inode value to avoid infinite recursion
 * endif
 * return -1 if error
 * return 0 if success, and no additional output is necessary
 * return 1 if success, and additional output is necessary
 *
 * Note: instead of having to deal with a partial directory within this
 * function, the function just sets fileState->deleteMe to to, which causes
 * a delStateDataDir to be run on it when the fileState is closed.
 * (note that closeStateDataDir should be run, even if this function 
 * fails...)
 */
int refreshState(struct stateDirHandle *fileState) {
	int result;
	struct stat fileStats;

	assert(fileState);
	assert(fileState->localDataFile);
	assert(fileState->localDataFileName);
	assert(fileState->dataFileName);
	assert(fileState->dataFileInode != 0);

	/*If there isn't a new file to pick up, just delete the state...*/
	if(fileState->dataFileInode == -1) {
			fileState->deleteMe = 1;
			return 0;
	}

	/*Otherwise, attempt to re-init the directory*/
	result = unlink(fileState->localDataFileName);
	if(result < 0) {
		perror("Couldn't unlink local data file");
		fileState->deleteMe = 1;
		return -1;
	}

	result = doLink(fileState->dataFileName, fileState->localDataFileName);

	if(result < 0) {
		fprintf(stderr, "Couldn't re-establish local data link\n");
		fileState->deleteMe = 1;
		return -1;
	}

	result = fclose(fileState->localDataFile);
	/*reset to flag value*/
	fileState->localDataFile = NULL;

	if(result < 0) {
		perror("Couldn't close local data file...");
		fileState->deleteMe = 1;
		return -1;
	}

	fileState->filePos = 0;

	if((fileState->localDataFile = fopen(fileState->localDataFileName, "r")) == NULL) {
		perror("Can't open the local data file");
		fileState->deleteMe = 1;
		return (-1);
	}

	result = stat(fileState->localDataFileName, &fileStats);
	if(result < 0) {
		perror("Can't stat the local data file");
		fileState->deleteMe = 1;
		return -1;
	}

	fileState->localDataFileInode = fileStats.st_ino;

	/*now let's make sure it hasn't changed already...*/
	if(fileState->localDataFileInode != fileState->dataFileInode) {
		fprintf(stderr, "File has changed during operation\n");
		fileState->dataFileInode = fileState->localDataFileInode;
	}

	return 1;
}

/* This function deletes the state directory.  It does _not_
 * close, spindle or mutilate the fileState handle.  It just
 * trashes the directory.
 * It:
 * removes the state file
 * removes the local data file (if present)
 * removes the new directory
 */
int delStateDataDir(struct stateDirHandle *fileState) {
	int result;
	int returnValue;

	returnValue = 0;

	result = unlink(fileState->stateFileName);
	if((result < 0) && (errno != ENOENT)) {
		perror("Can't unlink the state file");
		returnValue = -1;
	}

	result = unlink(fileState->localDataFileName);
	if((result < 0) && (errno != ENOENT)) {
		perror("Can't unlink the local data file");
		returnValue = -1;
	}

	result = rmdir(fileState->stateDataDirName);
	if((result < 0) && (errno != ENOENT)) {
		perror("Can't rmdir the state directory");
		returnValue = -1;
	}
	
	return returnValue;
}


/*This is a debugging function only
 *it is not used in the normal code base
 *It:
 *outputs all the current fileState values
 */
void printState(struct stateDirHandle *fileState) {
	fprintf(stderr, "fileState in review:\n");
	fprintf(stderr, "stateDirName: %s\n", fileState->stateDirName);
	fprintf(stderr, "dataFileName: %s\n", fileState->dataFileName);
	fprintf(stderr, "stateDirLockFileName: %s\n", fileState->stateDirLockFileName);
	fprintf(stderr, "stateDataDirName: %s\n", fileState->stateDataDirName);
	fprintf(stderr, "localDataFileName: %s\n", fileState->localDataFileName);
	fprintf(stderr, "stateFileName: %s\n", fileState->stateFileName);
	fprintf(stderr, "lock fd: %d\n", fileState->stateDirLockFile);
	fprintf(stderr, "localDataFileInode: %ld\n", fileState->localDataFileInode);
	fprintf(stderr, "dataFileInode: %ld\n", fileState->dataFileInode);
	fprintf(stderr, "filePos: %ld\n", fileState->filePos);
	fprintf(stderr, "\n\n");
}
