/* The Ketchup program
 * A program to efficiently get recent logs
 * Copyright (C) 2001
 * Joshua Hill <josh-ketchup@untruth.org>
 * License: GPL
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include "util.h"

/* This function gets complete lines from the data file
 * only get full lines.  if there is no full (LF terminated)
 * line available, make sure the file position is reset to the
 * beginning of the incomplete line.
 * Proviso: if the line is larger than (size - 2),
 * the first (size - 2) characters of the line are declared to be a 
 * complete line, and are LF terminated.
 * We do this to avoid situations where data is lost. (and instead
 * data is altered... certainly not ideal, but hopefully this is
 * a somewhat better failure mode...)
 * It:
 * gets the current file position
 * gets gets a line
 * if the line is not complete (LF terminated) then
 *    if the string is of length size-2 then
 *      terminate the string with a LF and a new NULL.
 *    else 
 *      set the file position back to the beginning of the line
 *    fi
 * endif
 */
char *getLine(char *buffer, size_t size, FILE *input) {
	long lastLine;
	char *strResult;
	int result;
	char lastChar;
	size_t lineLength;

	assert(buffer);
	assert(size > 0);
	assert(input);

	lastLine = ftell(input);
	if(lastLine < 0) {
		perror("Can't get file offset");
		return NULL;
	}

	/*Note that fgets reads in at most one less than the supplied
	 *characters, so in actuality the user string can only be
	 *of length (size-2)*/

	strResult = fgets(buffer, size - 1, input);
	if(strResult == NULL) {
		return NULL;
	}

	lineLength = strlen(buffer);

	lastChar = buffer[lineLength - 1];
	if(lastChar != 10) {
		/*was the line larger than the buffer?*/
		if(lineLength == (size - 2)) {
			fprintf(stderr, "long line detected.  Line broken up\n");

			/*Append a LF and re-terminate*/
			buffer[lineLength] = 10;
			buffer[lineLength + 1] = 0;
			lineLength++;
		} else {
			fprintf(stderr, "incomplete line...\n");
			result = fseek(input, lastLine, SEEK_SET);
			if(result < 0) {
				perror("Can't reposition file position");
				return NULL;
			}
			return NULL;
		}
	}
	
	return buffer;
}

/* This function finishes the file off, even if it is not a complete line.
 * This function is very similar to getLine
 * This function will modify the data by adding a linefeed if there is not
 * one present.
 * It:
 * gets a line using fgets
 * checks to see if it is in fact at the EOF
 * if not at EOF
 *    Output error message
 * endif
 * if line is not LF terminated
 *   LF terminate the line
 * fi
 */
 
char *finishFile(char *buffer, size_t size, FILE *input) {
	long curPos;
	long endPos;
	int result;
	int lineLength;

	char *strResult;

	assert(buffer);
	assert(size > 0);
	assert(input);

	strResult = fgets(buffer, size-1, input);
	if(strResult == NULL) {
		return NULL;
	}

	curPos = ftell(input);
	if(curPos < 0) {
		perror("Couldn't get current file position");
		return buffer;
	}

	result = fseek(input, 0, SEEK_END);
	if(result < 0) {
		perror("Couldn't set EOF location");
		return buffer;
	}

	endPos = ftell(input);
	if(endPos < 0) {
		perror("Couldn't get EOF location");
		return buffer;
	}
				
	if(curPos != endPos) {
		fprintf(stderr, "Unexpected file modification detected: Some data may be lost\n");
	}


	/*Check to see if the last line is complete...*/
	lineLength = strlen(buffer);

	if(buffer[lineLength - 1] != 10) {
		fprintf(stderr, "Last log line not LF terminated.\n");

		/*Append a LF and re-terminate*/
		buffer[lineLength] = 10;
		buffer[lineLength + 1] = 0;
		lineLength++;
	}
	
	return buffer;
}

/* This function does very little at this point.  It was originally
 * going to allow the user to select the number of significant directories.
 * that feature was abandoned, so now it just does some basic string
 * munging.
 * It:
 * duplicates the string
 * replaces all '/' characters with ':'
 * outputs the new string
 */
char *getBaseName(const char *filename) {
	char *outputString;
	char *curPos;

	assert(filename);

	outputString = strdup(filename);
	if(!outputString) {
		perror("Can't dup basename");
		return NULL;
	}

	curPos = outputString;

	while(*curPos) {
		if(*curPos == '/')
			*curPos = ':';
		curPos++;
	}

	return outputString;
}

/*This is a function that attempts to see if we are dealing with a
 *most-reduced path.  (That is, an absolute path, with no "/./" or
 *"/../" strings.  This is mainly an idiot catch... (the program _is_
 *intuitive, right!?!  WHAT YOU SAY!!!
 *
 *This function:
 *checks for a leading "/"
 *checks for the presence of the substring "/./"
 *checks for the presence of the substring "/../"
 */
int isAbsolutePath(const char *inputStr) {

	assert(inputStr);

	if(inputStr[0] != '/') {
		return 0;
	}

	if(strstr(inputStr, "/./")) {
		return 0;
	}

	if(strstr(inputStr, "/../")) {
		return 0;
	}

	return 1;
}

/*Unhappy library interaction.
 * strnlen is a GNU extension, so I'll implement it here...
 */
size_t my_strnlen(const char *s, size_t maxlen) {
	size_t cntr;

	for(cntr = 0; cntr < maxlen; cntr++) {
		if(s[cntr] == 0)
			break;
	}

	return cntr;
}
